// Copyright (c) 2026 Blacknon. All rights reserved.
// Use of this source code is governed by an MIT license
// that can be found in the LICENSE file.

fn main() {
    println!("{}", env!("CARGO_PKG_VERSION"));
}
